import datetime
import functools
import os
import uuid
from typing import Iterable

import requests.exceptions
from apscheduler.events import EVENT_JOB_ERROR
from apscheduler.executors.pool import ThreadPoolExecutor
from apscheduler.jobstores.memory import MemoryJobStore
from apscheduler.schedulers.blocking import BlockingScheduler
from sentry_sdk import capture_exception

from robotoff import settings, slack
from robotoff.insights.annotate import (
    UPDATED_ANNOTATION_RESULT,
    InsightAnnotatorFactory,
)
from robotoff.insights.importer import import_insights
from robotoff.metrics import (
    ensure_influx_database,
    save_facet_metrics,
    save_insight_metrics,
)
from robotoff.models import ProductInsight, db, with_db
from robotoff.prediction.category.matcher import predict_from_dataset
from robotoff.products import (
    CACHED_PRODUCT_STORE,
    Product,
    ProductDataset,
    fetch_dataset,
    has_dataset_changed,
)
from robotoff.utils import get_logger

from .latent import generate_quality_facets

settings.init_sentry()

logger = get_logger(__name__)


# Note: we do not use with_db, for atomicity is handled in annotator
def process_insights():
    with db.connection_context():
        processed = 0
        for insight in (
            ProductInsight.select()
            .where(
                ProductInsight.annotation.is_null(),
                ProductInsight.process_after.is_null(False),
                ProductInsight.process_after <= datetime.datetime.utcnow(),
            )
            .iterator()
        ):
            try:
                annotator = InsightAnnotatorFactory.get(insight.type)
                logger.info(
                    "Annotating insight %s (product: %s)", insight.id, insight.barcode
                )
                annotation_result = annotator.annotate(insight, 1, update=True)
                processed += 1

                if annotation_result == UPDATED_ANNOTATION_RESULT and insight.data.get(
                    "notify", False
                ):
                    slack.NotifierFactory.get_notifier().notify_automatic_processing(
                        insight
                    )
            except Exception as e:
                # continue to the next one
                # Note: annotator already rolled-back the transaction
                logger.exception(
                    f"exception {e} while handling annotation of insight %s (product) %s",
                    insight.id,
                    insight.barcode,
                )
    logger.info("%d insights processed", processed)


@with_db
def refresh_insights(with_deletion: bool = False):
    deleted = 0
    updated = 0
    product_store = CACHED_PRODUCT_STORE.get()

    datetime_threshold = datetime.datetime.utcnow().replace(
        hour=0, minute=0, second=0, microsecond=0
    )
    dataset_datetime = datetime.datetime.fromtimestamp(
        os.path.getmtime(settings.JSONL_MIN_DATASET_PATH)
    )

    if dataset_datetime.date() != datetime_threshold.date():
        logger.warning(
            "Dataset version is not up to date, aborting insight removal job"
        )
        return

    for insight in (
        ProductInsight.select()
        .where(
            ProductInsight.annotation.is_null(),
            ProductInsight.timestamp <= datetime_threshold,
            ProductInsight.server_domain == settings.OFF_SERVER_DOMAIN,
        )
        .iterator()
    ):
        product: Product = product_store[insight.barcode]

        if product is None:
            if with_deletion:
                # Product has been deleted from OFF
                logger.info("Product with barcode {} deleted".format(insight.barcode))
                deleted += 1
                insight.delete_instance()
        else:
            insight_updated = update_insight_attributes(product, insight)

            if insight_updated:
                updated += 1

    logger.info("{} insights deleted".format(deleted))
    logger.info("{} insights updated".format(updated))


def update_insight_attributes(product: Product, insight: ProductInsight) -> bool:
    to_update = False
    if insight.brands != product.brands_tags:
        logger.info(
            "Updating brand {} -> {} ({})".format(
                insight.brands, product.brands_tags, product.barcode
            )
        )
        to_update = True
        insight.brands = product.brands_tags

    if insight.countries != product.countries_tags:
        logger.info(
            "Updating countries {} -> {} ({})".format(
                insight.countries, product.countries_tags, product.barcode
            )
        )
        to_update = True
        insight.countries = product.countries_tags

    if insight.unique_scans_n != product.unique_scans_n:
        logger.info(
            "Updating unique scan count {} -> {} ({})".format(
                insight.unique_scans_n, product.unique_scans_n, product.barcode
            )
        )
        to_update = True
        insight.unique_scans_n = product.unique_scans_n

    if to_update:
        insight.save()

    return to_update


@with_db
def mark_insights():
    marked = 0
    for insight in (
        ProductInsight.select()
        .where(
            ProductInsight.automatic_processing == True,  # noqa: E712
            ProductInsight.process_after.is_null(),
            ProductInsight.annotation.is_null(),
        )
        .iterator()
    ):
        logger.info(
            "Marking insight {} as processable automatically "
            "(product: {})".format(insight.id, insight.barcode)
        )
        insight.process_after = datetime.datetime.utcnow() + datetime.timedelta(
            minutes=10
        )
        insight.save()
        marked += 1

    logger.info("{} insights marked".format(marked))
    return marked  # useful for tests


def _download_product_dataset():
    logger.info("Downloading new version of product dataset")

    if has_dataset_changed():
        fetch_dataset()


# this job does no use database
def _update_data():
    """Refreshes the PO product dump and updates the Elasticsearch index data."""
    try:
        _download_product_dataset()
    except requests.exceptions.RequestException:
        logger.exception("Exception during product dataset refresh")


def generate_insights():
    """Generate and import category insights from the latest dataset dump, for
    products added at day-1."""
    logger.info("Generating new category insights")

    datetime_threshold = datetime.datetime.utcnow().replace(
        hour=0, minute=0, second=0, microsecond=0
    ) - datetime.timedelta(days=1)
    dataset = ProductDataset(settings.JSONL_DATASET_PATH)
    product_predictions_iter = predict_from_dataset(dataset, datetime_threshold)

    with db:
        imported = import_insights(
            product_predictions_iter, server_domain=settings.OFF_SERVER_DOMAIN
        )
    logger.info("%s category insights imported", imported)


def transform_insight_iter(insights_iter: Iterable[dict]):
    for insight in insights_iter:
        for field, value in insight.items():
            if isinstance(value, uuid.UUID):
                insight[field] = str(value)
            elif isinstance(value, datetime.datetime):
                insight[field] = value.isoformat()

        yield insight


def exception_listener(event):
    if event.exception:
        capture_exception(event.exception)


# The scheduler is responsible for scheduling periodic work that Robotoff needs to perform.
def run():
    # ensure influxdb database exists
    ensure_influx_database()

    # This call needs to happen on every start of the scheduler to ensure we're not in
    # the state where Robotoff is unable to perform tasks because of missing data.
    _update_data()

    scheduler = BlockingScheduler()
    scheduler.add_executor(ThreadPoolExecutor(20))
    scheduler.add_jobstore(MemoryJobStore())

    # This job takes all of the newly added automatically-processable insights and sets the process_after field on them,
    # indicating when these insights should be auto-applied.
    scheduler.add_job(mark_insights, "interval", minutes=2, max_instances=1, jitter=20)

    # This job applies all of the automatically-processable insights that have not been applied yet.
    scheduler.add_job(
        process_insights, "interval", minutes=2, max_instances=1, jitter=20
    )

    # This job exports daily product metrics for monitoring.
    scheduler.add_job(save_facet_metrics, "cron", day="*", hour=1, max_instances=1)
    scheduler.add_job(save_insight_metrics, "cron", day="*", hour=1, max_instances=1)

    # This job refreshes data needed to generate insights.
    scheduler.add_job(_update_data, "cron", day="*", hour="3", max_instances=1)

    # This job updates the product insights state with respect to the latest PO dump by:
    # - Deleting non-annotated insights for deleted products and insights that
    #   are no longer applicable.
    # - Updating insight attributes.
    scheduler.add_job(
        functools.partial(refresh_insights, with_deletion=True),
        "cron",
        day="*",
        hour="4",
        max_instances=1,
    )

    # This job generates category insights using ElasticSearch from the last Product Opener data dump.
    scheduler.add_job(
        generate_insights, "cron", day="*", hour="4", minute=15, max_instances=1
    )

    scheduler.add_job(
        generate_quality_facets,
        "cron",
        day="*",
        hour="5",
        minute=25,
        max_instances=1,
    )

    scheduler.add_listener(exception_listener, EVENT_JOB_ERROR)
    scheduler.start()



# This package describes the Postgres tables Robotoff is writing to.
import datetime
import functools
import uuid
from typing import Iterable

import peewee
from playhouse.postgres_ext import BinaryJSONField, PostgresqlExtDatabase
from playhouse.shortcuts import model_to_dict

from robotoff import settings

db = PostgresqlExtDatabase(
    settings.POSTGRES_DB,
    user=settings.POSTGRES_USER,
    password=settings.POSTGRES_PASSWORD,
    host=settings.POSTGRES_HOST,
    port=5432,
    autoconnect=False,
)


def with_db(fn):
    """Run function inside a SQL transaction."""

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        with db:
            # use atomic transaction to avoid falling in a bad state
            # (error in the main transaction)
            return fn(*args, **kwargs)

    return wrapper


def batch_insert(model_cls, data: Iterable[dict], batch_size=100) -> int:
    rows = 0
    inserts = []

    for item in data:
        inserts.append(item)
        rows += 1

        if rows % batch_size == 0:
            model_cls.insert_many(inserts).execute()
            inserts = []

    if inserts:
        model_cls.insert_many(inserts).execute()

    return rows


def crop_image_url(source_image, bounding_box) -> str:
    base_url = settings.OFF_IMAGE_BASE_URL + source_image
    y_min, x_min, y_max, x_max = bounding_box
    base_robotoff_url = settings.BaseURLProvider().robotoff().get()
    return f"{base_robotoff_url}/api/v1/images/crop?image_url={base_url}&y_min={y_min}&x_min={x_min}&y_max={y_max}&x_max={x_max}"


class BaseModel(peewee.Model):
    class Meta:
        database = db
        legacy_table_names = False

    def to_dict(self, **kwargs):
        return model_to_dict(self, **kwargs)


class ProductInsight(BaseModel):
    """Table to store the insights generated by Robotoff."""

    # ID is the unique ID for this insight.
    id = peewee.UUIDField(primary_key=True, default=uuid.uuid4)

    # Barcode represents the barcode of the product for which the insight was generated.
    barcode = peewee.CharField(max_length=100, null=False, index=True)

    # Type represents the insight type - must match one of the types in robotoff.insights.dataclass.InsightType.
    type = peewee.CharField(max_length=256, index=True)

    # Contains some additional data based on the type of the insight from above.
    # NOTE: there is no 1:1 mapping between the type and the JSON format provided here, for example for
    # type==label, the data here could be: {"logo_id":X,"confidence":Y}, or {"text":X,"notify":Y}
    data = BinaryJSONField(index=True, default=dict)

    # Timestamp is the timestamp of when this insight was imported into the DB.
    timestamp = peewee.DateTimeField(null=True, index=True)

    # Stores the timestamp of when this insight was annotated, either by a human annotator or automatically.
    completed_at = peewee.DateTimeField(null=True)

    # The annotation of the given insight. Four possible values are possible:
    # null = This insight has not been annotated
    #  -1 = Rejected
    # 0 = 'I don't know'
    # 1 = Validated
    annotation = peewee.IntegerField(null=True, index=True)

    # Saves the value returned by Annotator.annotate
    # The value is mapped at `AnnotationStatus(enum)` class in annotate.py
    annotated_result = peewee.IntegerField(null=True, index=False)

    # The number of votes for this annotation.
    # Stored here for quick sorting.
    n_votes = peewee.IntegerField(null=False, index=True)

    # If the insight was annotated manually, this field stores the username of the annotator
    # (or first annotator, if multiple votes were cast).
    username = peewee.TextField(index=True, null=True)

    # Stores the list of counties that are associated with the product.
    # E.g. possible values are "en:united-states" or "en:france".
    countries = BinaryJSONField(null=True, index=True)

    # Stores the list of brands that are associated with the product.
    brands = BinaryJSONField(null=True, index=True)

    # Specifies the timestamp on an automatic_processing insight after which the insight should be applied.
    process_after = peewee.DateTimeField(null=True)

    # Value tag contains the normalized 'value' of the insight based on the OFF taxonomies.
    # Some insight types will populate both this field and the 'value' field below, some only this field.
    value_tag = peewee.TextField(null=True, index=True)

    # Value stores the value of the insight, for example the brand name for a product or the numeric value of the weight of the product.
    # Some insight types will populate both this field and the 'value_tag' field above, some only this field.
    # This field is set for the following insight types: brand, expiration_date, packager_code, packaging, product_weight and store.
    value = peewee.TextField(null=True, index=True)

    # If the insight was based on an image, this field points to that image.
    # The complete image path can be constructed with robotoff.settings.OFF_IMAGE_BASE_URL + source_image.
    source_image = peewee.TextField(null=True, index=True)

    # Automatic processing is set on insights where Robotoff deems the prediction to be valid, e.g. if the prediction is above a
    # pre-determined threshold.
    automatic_processing = peewee.BooleanField(default=False, index=True)

    server_domain = peewee.TextField(
        null=True, help_text="server domain linked to the insight", index=True
    )
    server_type = peewee.CharField(
        null=True,
        max_length=10,
        help_text="project associated with the server_domain, "
        "one of 'off', 'obf', 'opff', 'opf'",
        index=True,
    )

    # This field refers to the number of unique IPs that have scanned this product.
    unique_scans_n = peewee.IntegerField(default=0, index=True)

    # A reserved barcode is a barcode that is a 'per-weight' barcode.
    reserved_barcode = peewee.BooleanField(default=False, index=True)

    # Predictor stores what ML model/OCR processing generated this insight.
    predictor = peewee.CharField(max_length=100, null=True, index=True)

    # annotation campaigns enable contributors to focus their efforts (on
    # Hunger Games) on a subset of products. Each product have 0+ campaign
    # tags
    campaign = BinaryJSONField(null=True, index=True, default=list)


class Prediction(BaseModel):
    barcode = peewee.CharField(max_length=100, null=False, index=True)
    type = peewee.CharField(max_length=256, index=True)
    data = BinaryJSONField(index=True)
    timestamp = peewee.DateTimeField(index=True)
    value_tag = peewee.TextField(null=True)
    value = peewee.TextField(null=True)
    source_image = peewee.TextField(null=True, index=True)
    automatic_processing = peewee.BooleanField(null=True)
    server_domain = peewee.TextField(
        help_text="server domain linked to the insight", index=True
    )
    predictor = peewee.CharField(max_length=100, null=True)


class AnnotationVote(BaseModel):
    id = peewee.UUIDField(primary_key=True, default=uuid.uuid4)
    # The insight this vote belongs to.
    insight_id = peewee.ForeignKeyField(
        ProductInsight, null=False, backref="votes", on_delete="CASCADE"
    )
    # The username of the voter - if logged in.
    # Currently logged in users do not need to vote on an insight for their annotation to
    # be applied.
    username = peewee.TextField(index=True, null=True)
    # The value of the annotation, see ProductInsight.annotation.
    value = peewee.IntegerField(
        null=False, choices=[(-1, "False"), (0, "Unknown"), (1, "True")]
    )
    # If the request has a device_id, use that - otherwise use a secure hash of the IP address.
    device_id = peewee.TextField(index=True, null=False)
    # Creation date for bookkeeping.
    timestamp = peewee.DateTimeField(null=False, default=datetime.datetime.utcnow)


class ImageModel(BaseModel):
    barcode = peewee.CharField(max_length=100, null=False, index=True)
    uploaded_at = peewee.DateTimeField(null=True, index=True)
    image_id = peewee.CharField(max_length=50, null=False, index=True)
    # The complete image path can be constructed with robotoff.settings.OFF_IMAGE_BASE_URL + source_image.
    source_image = peewee.TextField(null=False, index=True)
    width = peewee.IntegerField(null=False, index=True)
    height = peewee.IntegerField(null=False, index=True)
    deleted = peewee.BooleanField(null=False, index=True, default=False)
    server_domain = peewee.TextField(null=True, index=True)
    server_type = peewee.CharField(null=True, max_length=10, index=True)

    class Meta:
        table_name = "image"


class ImagePrediction(BaseModel):
    """Table to store computer vision predictions (object detection,
    image segmentation,...) made by custom models.

    They are created by API `ImagePredictorResource`, `ImagePredictionImporterResource`
    or cli `import_logos`

    Predictions come from a model, see `ObjectDetectionModel` in
    robotoff/types.py for available models.
    """

    type = peewee.CharField(max_length=256)
    model_name = peewee.CharField(max_length=100, null=False, index=True)
    model_version = peewee.CharField(max_length=256, null=False, index=True)
    data = BinaryJSONField(index=True)
    timestamp = peewee.DateTimeField(null=True)
    image = peewee.ForeignKeyField(ImageModel, null=False, backref="predictions")
    max_confidence = peewee.FloatField(
        null=True,
        index=True,
        help_text="for object detection models, confidence of the highest confident"
        "object detected, null if no object was detected",
    )


class LogoAnnotation(BaseModel):
    """Annotation(s) for an image prediction
    (an image prediction might lead to several annotations)

    At the moment, this is mostly for logo (see run_logo_object_detection),
    when we have a logo prediction above a certain threshold we create an entry,
    to ask user for annotation on the logo (https://hunger.openfoodfacts.org/logos)
    and eventual annotation will land there.
    """

    image_prediction = peewee.ForeignKeyField(
        ImagePrediction, null=False, backref="logo_detections"
    )
    index = peewee.IntegerField(null=False, constraints=[peewee.Check("index >= 0")])
    bounding_box = BinaryJSONField(null=False)
    score = peewee.FloatField(null=False)
    annotation_value = peewee.CharField(null=True, index=True)
    annotation_value_tag = peewee.CharField(null=True, index=True)
    taxonomy_value = peewee.CharField(null=True, index=True)
    annotation_type = peewee.CharField(
        null=True,
        index=True,
        help_text="Category of the annotation. Current possible values: brand, "
        "category, label, no_logo, nutrition_label, packager_code, packaging, "
        "qr_code, store",
    )
    username = peewee.TextField(null=True, index=True)
    completed_at = peewee.DateTimeField(null=True, index=True)
    nearest_neighbors = BinaryJSONField(null=True)

    class Meta:
        constraints = [peewee.SQL("UNIQUE(image_prediction_id, index)")]

    def get_crop_image_url(self) -> str:
        return crop_image_url(
            self.image_prediction.image.source_image, self.bounding_box
        )


class LogoConfidenceThreshold(BaseModel):
    type = peewee.CharField(null=True, index=True)
    value = peewee.CharField(null=True, index=True)
    threshold = peewee.FloatField(null=False)


MODELS = [
    Prediction,
    ProductInsight,
    ImageModel,
    ImagePrediction,
    LogoAnnotation,
    LogoConfidenceThreshold,
    AnnotationVote,
]



import abc
import dataclasses

from langid import langid

from robotoff.utils.cache import CachedStore


@dataclasses.dataclass
class LanguagePrediction:
    lang: str
    confidence: float


class LanguageIdentifier(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def predict(
        self, text: str, k: int = 10, threshold: float = 0.0
    ) -> list[LanguagePrediction]:
        pass


class LangidLanguageIdentifier(LanguageIdentifier):
    def __init__(self, model):
        self.model = model

    @classmethod
    def load(cls):
        model = langid.LanguageIdentifier.from_modelstring(
            langid.model, norm_probs=True
        )
        return cls(model)

    def predict(
        self, text: str, k: int = 10, threshold: float = 0.0
    ) -> list[LanguagePrediction]:
        predictions: list[LanguagePrediction] = []
        rank = self.model.rank(text)

        added: int = 0
        for language, confidence in rank:
            if confidence >= threshold:
                prediction = LanguagePrediction(language, confidence)
                predictions.append(prediction)
                added += 1
            else:
                break

            if added >= k:
                break

        return predictions


class FastTextLanguageIdentifier(LanguageIdentifier):
    def __init__(self, model):
        self.model = model

    def predict(
        self, text: str, k: int = 10, threshold: float = 0.0
    ) -> list[LanguagePrediction]:
        predictions: list[LanguagePrediction] = []
        languages, confidences = self.model.predict(text, k=k, threshold=threshold)

        for language, confidence in zip(languages, confidences):
            # language str format is __label__en
            language = language[9:]
            prediction = LanguagePrediction(language, confidence)
            predictions.append(prediction)

        return predictions


DEFAULT_LANGUAGE_IDENTIFIER = CachedStore(LangidLanguageIdentifier.load)



import re
from typing import Union

from robotoff.prediction.types import Prediction
from robotoff.types import PredictionType
from robotoff.utils.types import JSONType

from .dataclass import OCRField, OCRRegex, OCRResult, get_text

EXTRACTOR_VERSION = "2"


NutrientMentionType = tuple[str, list[str]]

NUTRIENT_MENTION: dict[str, list[NutrientMentionType]] = {
    "energy": [
        ("[ée]nergie", ["fr", "de"]),
        ("valeurs? [ée]nerg[ée]tiques?", ["fr"]),
        ("energy", ["en"]),
        ("calories", ["fr", "en"]),
        ("energia", ["es", "it", "pt"]),
        ("valor energ[ée]tico", ["es"]),
    ],
    "saturated_fat": [
        ("mati[èe]res? grasses? satur[ée]s?", ["fr"]),
        ("acides? gras satur[ée]s?", ["fr"]),
        ("dont satur[ée]s?", ["fr"]),
        ("acidi grassi saturi", ["it"]),
        ("saturated fat", ["en"]),
        ("of which saturates", ["en"]),
        ("verzadigde vetzuren", ["nl"]),
        ("waarvan verzadigde", ["nl"]),
        ("gesättigte fettsäuren", ["de"]),
        ("[aá]cidos grasos saturados", ["es"]),
    ],
    "trans_fat": [("mati[èe]res? grasses? trans", ["fr"]), ("trans fat", ["en"])],
    "fat": [
        ("mati[èe]res? grasses?", ["fr"]),
        ("graisses?", ["fr"]),
        ("lipides?", ["fr"]),
        ("total fat", ["en"]),
        ("vetten", ["nl"]),
        ("fett", ["de"]),
        ("grasas", ["es"]),
        ("grassi", ["it"]),
        ("l[íi]pidos", ["es"]),
    ],
    "sugar": [
        ("sucres?", ["fr"]),
        ("sugars?", ["en"]),
        ("zuccheri", ["it"]),
        ("suikers?", ["nl"]),
        ("zucker", ["de"]),
        ("az[úu]cares", ["es"]),
    ],
    "carbohydrate": [
        ("total carbohydrate", ["en"]),
        ("glucids?", ["fr"]),
        ("glucides?", ["en"]),
        ("carboidrati", ["it"]),
        ("koolhydraten", ["nl"]),
        ("koolhydraat", ["nl"]),
        ("kohlenhydrate", ["de"]),
        ("hidratos de carbono", ["es"]),
    ],
    "protein": [
        ("prot[ée]ines?", ["fr"]),
        ("protein", ["en"]),
        ("eiwitten", ["nl"]),
        ("eiweiß", ["de"]),
        ("prote[íi]nas", ["es"]),
    ],
    "salt": [
        ("sel", ["fr"]),
        ("salt", ["en"]),
        ("zout", ["nl"]),
        ("salz", ["de"]),
        ("sale", ["it"]),
        ("sal", ["es"]),
    ],
    "fiber": [
        ("fibres?", ["en", "fr", "it"]),
        ("fibers?", ["en"]),
        ("fibres? alimentaires?", ["fr"]),
        ("(?:voedings)?vezels?", ["nl"]),
        ("ballaststoffe", ["de"]),
        ("fibra(?: alimentaria)?", ["es"]),
    ],
    "nutrition_values": [
        ("informations? nutritionnelles?(?: moyennes?)?", ["fr"]),
        ("valeurs? nutritionnelles?(?: moyennes?)?", ["fr"]),
        ("analyse moyenne pour", ["fr"]),
        ("valeurs? nutritives?", ["fr"]),
        ("valeurs? moyennes?", ["fr"]),
        ("nutrition facts?", ["en"]),
        ("average nutritional values?", ["en"]),
        ("valori nutrizionali medi", ["it"]),
        ("gemiddelde waarden per", ["nl"]),
    ],
}


NUTRIENT_UNITS: dict[str, list[str]] = {
    "energy": ["kj", "kcal"],
    "saturated_fat": ["g"],
    "trans_fat": ["g"],
    "fat": ["g"],
    "sugar": ["g"],
    "carbohydrate": ["g"],
    "protein": ["g"],
    "salt": ["g"],
    "fiber": ["g"],
}


def generate_nutrient_regex(
    nutrient_mentions: list[NutrientMentionType], units: list[str]
):
    nutrient_names = [x[0] for x in nutrient_mentions]
    nutrient_names_str = "|".join(nutrient_names)
    units_str = "|".join(units)
    return re.compile(
        r"(?<!\w)({}) ?(?:[:-] ?)?([0-9]+[,.]?[0-9]*) ?({})(?!\w)".format(
            nutrient_names_str, units_str
        )
    )


def generate_nutrient_mention_regex(nutrient_mentions: list[NutrientMentionType]):
    sub_re = "|".join(
        r"(?P<{}>{})".format("{}_{}".format("_".join(lang), i), name)
        for i, (name, lang) in enumerate(nutrient_mentions)
    )
    return re.compile(r"(?<!\w){}(?!\w)".format(sub_re))


NUTRIENT_VALUES_REGEX = {
    nutrient: OCRRegex(
        generate_nutrient_regex(NUTRIENT_MENTION[nutrient], units),
        field=OCRField.full_text_contiguous,
        lowercase=True,
    )
    for nutrient, units in NUTRIENT_UNITS.items()
}

NUTRIENT_MENTIONS_REGEX: dict[str, OCRRegex] = {
    nutrient: OCRRegex(
        generate_nutrient_mention_regex(NUTRIENT_MENTION[nutrient]),
        field=OCRField.full_text_contiguous,
        lowercase=True,
    )
    for nutrient in NUTRIENT_MENTION
}


def find_nutrient_values(content: Union[OCRResult, str]) -> list[Prediction]:
    nutrients: JSONType = {}

    for regex_code, ocr_regex in NUTRIENT_VALUES_REGEX.items():
        text = get_text(content, ocr_regex)

        if not text:
            continue

        for match in ocr_regex.regex.finditer(text):
            value = match.group(2).replace(",", ".")
            unit = match.group(3)
            nutrients.setdefault(regex_code, [])
            nutrients[regex_code].append(
                {
                    "raw": match.group(0),
                    "nutrient": regex_code,
                    "value": value,
                    "unit": unit,
                }
            )

    if not nutrients:
        return []

    return [
        Prediction(
            type=PredictionType.nutrient,
            data={"nutrients": nutrients, "version": EXTRACTOR_VERSION},
        )
    ]


def find_nutrient_mentions(content: Union[OCRResult, str]) -> list[Prediction]:
    nutrients: JSONType = {}

    for regex_code, ocr_regex in NUTRIENT_MENTIONS_REGEX.items():
        text = get_text(content, ocr_regex)

        if not text:
            continue

        for match in ocr_regex.regex.finditer(text):
            nutrients.setdefault(regex_code, [])
            group_dict = {k: v for k, v in match.groupdict().items() if v is not None}

            languages: list[str] = []
            if group_dict:
                languages_raw = list(group_dict.keys())[0]
                languages = languages_raw.rsplit("_", maxsplit=1)[0].split("_")

            nutrients[regex_code].append(
                {
                    "raw": match.group(0),
                    "span": list(match.span()),
                    "languages": languages,
                }
            )

    if not nutrients:
        return []

    return [
        Prediction(
            type=PredictionType.nutrient_mention,
            data={"mentions": nutrients, "version": EXTRACTOR_VERSION},
        )
    ]



import pathlib

from robotoff import settings
from robotoff.off import generate_image_url
from robotoff.products import ProductDataset
from robotoff.utils import get_logger, http_session
from robotoff.utils.types import JSONType

logger = get_logger()

JSONL_SHUF_DATASET_PATH = settings.DATASET_DIR / "products-shuf.jsonl.gz"
ds = ProductDataset(JSONL_SHUF_DATASET_PATH)
IMAGE_DATASET_DIR = settings.PROJECT_DIR / "image_dataset"
NUTRITION_TABLE_IMAGE_DIR = IMAGE_DATASET_DIR / "nutrition-table-2"


def load_seen_set() -> set[str]:
    seen_set = set()

    with open(IMAGE_DATASET_DIR / "dataset.txt") as f:
        for line in f:
            if line:
                line = line.strip("\n")
                barcode, _ = line.split("_")
                seen_set.add(barcode)

    return seen_set


def save_image(
    directory: pathlib.Path, image_meta: JSONType, barcode: str, override: bool = False
):
    image_id = image_meta["imgid"]
    image_name = "{}_{}.jpg".format(barcode, image_id)
    image_path = directory / image_name

    if image_path.exists() and not override:
        return

    image_url = generate_image_url(barcode, image_id)
    logger.info("Downloading image {}".format(image_url))
    r = http_session.get(image_url)

    with open(str(image_path), "wb") as fd:
        logger.info("Saving image in {}".format(image_path))
        for chunk in r.iter_content(chunk_size=128):
            fd.write(chunk)


seen_set = load_seen_set()
count = 0

for product in (
    ds.stream()
    .filter_by_state_tag("en:complete")
    .filter_by_country_tag("en:france")
    .filter_nonempty_text_field("code")
    .filter_nonempty_tag_field("images")
):
    barcode = product["code"]

    if barcode in seen_set:
        print("Product already seen: {}".format(barcode))
        continue

    has_nutrition = False
    has_front = False

    for image_key, image_meta in product.get("images", {}).items():
        if not has_nutrition and image_key.startswith("nutrition"):
            has_nutrition = True
            save_image(NUTRITION_TABLE_IMAGE_DIR, image_meta, barcode)
            count += 1
            continue

        elif not has_front and image_key.startswith("front"):
            has_front = True
            save_image(NUTRITION_TABLE_IMAGE_DIR, image_meta, barcode)
            count += 1
            continue

    if count >= 600:
        print("Breaking")
        print(count)
        break



import datetime
from typing import Optional
from urllib.parse import urlparse

import requests
from influxdb_client import InfluxDBClient
from influxdb_client.client.write_api import SYNCHRONOUS
from peewee import fn

from robotoff import settings
from robotoff.models import ProductInsight, with_db
from robotoff.utils import get_logger, http_session

logger = get_logger(__name__)

URL_PATHS: list[str] = [
    "/ingredients-analysis?json=1",
    "/data-quality?json=1",
    "/ingredients?stats=1&json=1",
    "/states?json=1",
    "/misc?json=1",
]

COUNTRY_TAGS = [
    "world",
    "us",
    "uk",
    "fr",
    "es",
    "it",
    "be",
    "nl",
    "de",
    "ch",
    "be",
    "ca",
    "au",
    "mx",
    "at",
    "ie",
    "pl",
    "pt",
    "se",
    "ru",
    "th",
    "ma",
    "lu",
    "re",
    "ro",
    "bg",
    "hu",
    "dz",
    "dk",
    "br",
    "cz",
    "sg",
    "fi",
    "ar",
    "gd",
    "jp",
    "no",
    "in",
    "tn",
    "dk",
]


def get_influx_client() -> Optional[InfluxDBClient]:
    if not settings.INFLUXDB_HOST:
        return None
    return InfluxDBClient(
        url=f"http://{settings.INFLUXDB_HOST}:{settings.INFLUXDB_PORT}",
        token=settings.INFLUXDB_AUTH_TOKEN,
        org=settings.INFLUXDB_ORG,
    )


def ensure_influx_database():
    client = get_influx_client()
    if client is not None:
        try:
            bucket_client = client.buckets_api()
            bucket_names = [
                bucket.name for bucket in bucket_client.find_buckets().buckets
            ]
            if settings.INFLUXDB_BUCKET not in bucket_names:
                # create it
                client.bucket_name(bucket_name=settings.INFLUXDB_BUCKET)
                logger.warning(
                    "Creating influxdb bucket %r as it does not exist yet",
                    settings.INFLUXDB_BUCKET,
                )
        except Exception:
            # better be fail safe, our job is not that important !
            logger.exception("Error on ensure_influx_database")


def get_product_count(country_tag: str) -> int:
    r = http_session.get(
        settings.BaseURLProvider().country(country_tag).get() + "/3.json?fields=null",
        auth=settings._off_request_auth,
    ).json()
    return int(r["count"])


def save_facet_metrics():
    inserts = []
    target_datetime = datetime.datetime.now()
    product_counts = {
        country_tag: get_product_count(country_tag) for country_tag in COUNTRY_TAGS
    }

    for country_tag in COUNTRY_TAGS:
        count = product_counts[country_tag]

        for url_path in URL_PATHS:
            inserts += generate_metrics_from_path(
                country_tag, url_path, target_datetime, count
            )

        inserts += generate_metrics_from_path(
            country_tag,
            "/entry-date/{}/contributors?json=1".format(
                # get contribution metrics for the previous day
                (target_datetime - datetime.timedelta(days=1)).strftime("%Y-%m-%d")
            ),
            target_datetime,
            facet="contributors",
        )

    inserts += generate_metrics_from_path("world", "/countries?json=1", target_datetime)
    client = get_influx_client()
    if client is not None:
        write_client = client.write_api(write_options=SYNCHRONOUS)
        write_client.write(bucket=settings.INFLUXDB_BUCKET, record=inserts)


def get_facet_name(url: str) -> str:
    return urlparse(url)[2].strip("/").replace("-", "_")


def generate_metrics_from_path(
    country_tag: str,
    path: str,
    target_datetime: datetime.datetime,
    count: Optional[int] = None,
    facet: Optional[str] = None,
) -> list[dict]:
    inserts: list[dict] = []
    url = settings.BaseURLProvider().country(country_tag + "-en").get() + path

    if facet is None:
        facet = get_facet_name(url)

    try:
        r = http_session.get(url, timeout=60, auth=settings._off_request_auth)
    except requests.exceptions.Timeout:
        logger.error("OFF request timeout (60s): {}".format(url))
        return inserts

    if not r.ok:
        logger.error("Error during OFF request: {}".format(r.status_code))
        return inserts

    try:
        data = r.json()
    except requests.exceptions.JSONDecodeError as e:
        logger.error("Error during OFF request JSON decoding:\n{}".format(e))
        return inserts

    for tag in data["tags"]:
        name = tag["name"]
        products = tag["products"]
        fields = {"products": products}

        if "percent" in tag:
            fields["percent"] = float(tag["percent"])

        elif count is not None:
            fields["percent"] = products * 100 / count

        tag_id = tag["id"]

        inserts.append(
            {
                "measurement": "facets",
                "tags": {
                    "tag_name": name,
                    "tag_id": tag_id,
                    "country": country_tag,
                    "facet": facet,
                },
                "time": target_datetime.isoformat(),
                "fields": fields,
            }
        )
    return inserts


def save_insight_metrics():
    """Save number of insights, grouped by the following fields:
    - type
    - annotation
    - automatic_processing
    - predictor
    - reserved_barcode
    """
    target_datetime = datetime.datetime.now()

    if (client := get_influx_client()) is not None:
        write_client = client.write_api(write_options=SYNCHRONOUS)
        inserts = generate_insight_metrics(target_datetime)
        write_client.write(bucket=settings.INFLUXDB_BUCKET, record=inserts)


@with_db
def generate_insight_metrics(target_datetime: datetime.datetime) -> list[dict]:
    group_by_fields = [
        ProductInsight.type,
        ProductInsight.annotation,
        ProductInsight.automatic_processing,
        ProductInsight.predictor,
        ProductInsight.reserved_barcode,
    ]
    inserts = []
    query_results = (
        ProductInsight.select(
            *group_by_fields,
            fn.COUNT(ProductInsight.id).alias("count"),
        )
        .group_by(*group_by_fields)
        .dicts()
    )
    total_count = sum(query_result["count"] for query_result in query_results)

    for query_result in query_results:
        count = query_result.pop("count")
        inserts.append(
            {
                "measurement": "insights",
                "tags": query_result,
                "time": target_datetime.isoformat(),
                "fields": {"count": count, "percent": count / total_count},
            }
        )
    return inserts



import threading
import time
from typing import Callable, Optional

from rq import Queue
from rq.job import Job

from robotoff.redis import redis_conn
from robotoff.types import WorkerQueue

high_queue = Queue(WorkerQueue.robotoff_high.value, connection=redis_conn)
low_queue = Queue(WorkerQueue.robotoff_low.value, connection=redis_conn)


def enqueue_in_job(
    func: Callable,
    queue: Queue,
    job_delay: float,
    job_kwargs: Optional[dict] = None,
    **kwargs
):
    """Enqueue a job in `job_delay` seconds.

    Launch a new Thread where we sleep `job_delay` seconds and the job is
    then enqueued.

    :param job_delay: number of seconds to sleep before sending the job to the
    queue
    """
    threading.Thread(
        target=_enqueue_in_job,
        args=(func, queue, job_delay, job_kwargs, kwargs),
    ).start()


def _enqueue_in_job(
    func: Callable,
    queue: Queue,
    job_delay: float,
    job_kwargs: Optional[dict],
    kwargs,
):
    time.sleep(job_delay)
    enqueue_job(func, queue, job_kwargs, **kwargs)


def enqueue_job(
    func: Callable, queue: Queue, job_kwargs: Optional[dict] = None, **kwargs
):
    """Create a new job from the function and kwargs and enqueue it in the
    queue.

    The function will be called by one of the rq workers. For safety, only
    keyword parameters can be provided to the function.

    :param func: the function to use
    :param queue: the queue to use
    :param job_kwargs: optional kwargs parameters to provide to `Job.create`
    """
    job_kwargs = job_kwargs or {}
    job = Job.create(func=func, kwargs=kwargs, connection=redis_conn, **job_kwargs)
    return queue.enqueue_job(job=job)
